@@ -1,48 +1,99 @@
-import os
+import errno
 import hashlib
 import json
-import random
-import stat
-import string
+import os
 import sys
 import tempfile
 import warnings
 
+from appdirs import AppDirs
+
 from rply.errors import ParserGeneratorError, ParserGeneratorWarning
 from rply.grammar import Grammar
 from rply.parser import LRParser
-from rply.utils import IdentityDict, Counter, iteritems, itervalues
+from rply.utils import Counter, IdentityDict, iteritems, itervalues
 
 
 LARGE_VALUE = sys.maxsize
 
 
 class ParserGenerator(object):
+    """
+    A ParserGenerator represents a set of production rules, that define a
+    sequence of terminals and non-terminals to be replaced with a non-terminal,
+    which can be turned into a parser.
+
+    :param tokens: A list of token (non-terminal) names.
+    :param precedence: A list of tuples defining the order of operation for
+                       avoiding ambiguity, consisting of a string defining
+                       associativity (left, right or nonassoc) and a list of
+                       token names with the same associativity and level of
+                       precedence.
+    :param cache_id: A string specifying an ID for caching.
+    """
     VERSION = 1
 
     def __init__(self, tokens, precedence=[], cache_id=None):
         self.tokens = tokens
         self.productions = []
         self.precedence = precedence
-        if cache_id is None:
-            # This ensures that we always go through the caching code.
-            cache_id = "".join(random.choice(string.ascii_letters) for _ in range(6))
         self.cache_id = cache_id
         self.error_handler = None
 
     def production(self, rule, precedence=None):
+        """
+        A decorator that defines one or many production rules and registers
+        the decorated function to be called with the terminals and
+        non-terminals matched by those rules.
+
+        A `rule` should consist of a name defining the non-terminal returned
+        by the decorated function and one or more sequences of pipe-separated
+        non-terminals and terminals that are supposed to be replaced::
+
+            replacing_non_terminal : TERMINAL1 non_term1 | TERMINAL2 non_term2
+
+        The name of the non-terminal replacing the sequence is on the left,
+        separated from the sequence by a colon. The whitespace around the colon
+        is required.
+
+        Knowing this we can define productions::
+
+            pg = ParserGenerator(['NUMBER', 'ADD'])
+
+            @pg.production('number : NUMBER')
+            def expr_number(p):
+                return BoxInt(int(p[0].getstr()))
+
+            @pg.production('expr : number ADD number')
+            def expr_add(p):
+                return BoxInt(p[0].getint() + p[2].getint())
+
+        If a state was passed to the parser, the decorated function is
+        additionally called with that state as first argument.
+        """
         parts = rule.split()
         production_name = parts[0]
         if parts[1] != ":":
             raise ParserGeneratorError("Expecting :")
-        syms = parts[2:]
+
+        body = " ".join(parts[2:])
+        prods = body.split("|")
 
         def inner(func):
-            self.productions.append((production_name, syms, func, precedence))
+            for production in prods:
+                syms = production.split()
+                self.productions.append((production_name, syms, func, precedence))
             return func
         return inner
 
     def error(self, func):
+        """
+        Sets the error handler that is called with the state (if passed to the
+        parser) and the token the parser errored on.
+
+        Currently error handlers must raise an exception. If an error handler
+        is not defined, a :exc:`rply.ParsingError` will be raised.
+        """
         self.error_handler = func
         return func
 
@@ -70,7 +121,9 @@ class ParserGenerator(object):
             "start": table.grammar.start,
             "terminals": sorted(table.grammar.terminals),
             "precedence": table.grammar.precedence,
-            "productions": [(p.name, p.prod, p.prec) for p in table.grammar.productions],
+            "productions": [
+                (p.name, p.prod, p.prec) for p in table.grammar.productions
+            ],
         }
 
     def data_is_valid(self, g, data):
@@ -123,49 +176,60 @@ class ParserGenerator(object):
         g.compute_first()
         g.compute_follow()
 
-        # win32 temp directories are already per-user
-        if os.name == "nt":
-            cache_file = os.path.join(
-                tempfile.gettempdir(),
-                "rply-%s-%s-%s.json" % (self.VERSION, self.cache_id, self.compute_grammar_hash(g))
-            )
-        else:
-            cache_file = os.path.join(
-                tempfile.gettempdir(),
-                "rply-%s-%s-%s-%s.json" % (self.VERSION, os.getuid(), self.cache_id, self.compute_grammar_hash(g))
-            )
         table = None
-        if os.path.exists(cache_file):
-            with open(cache_file) as f:
-                data = json.load(f)
-                stat_result = os.fstat(f.fileno())
-            if (
-                os.name == "nt" or (
-                    stat_result.st_uid == os.getuid() and
-                    stat.S_IMODE(stat_result.st_mode) == 0o0600
+        if self.cache_id is not None:
+            cache_dir = AppDirs("rply").user_cache_dir
+            cache_file = os.path.join(
+                cache_dir,
+                "%s-%s-%s.json" % (
+                    self.cache_id, self.VERSION, self.compute_grammar_hash(g)
                 )
-            ):
+            )
+
+            if os.path.exists(cache_file):
+                with open(cache_file) as f:
+                    data = json.load(f)
                 if self.data_is_valid(g, data):
                     table = LRTable.from_cache(g, data)
         if table is None:
             table = LRTable.from_grammar(g)
-            fd = os.open(cache_file, os.O_RDWR | os.O_CREAT | os.O_EXCL, 0o0600)
-            with os.fdopen(fd, "w") as f:
-                json.dump(self.serialize_table(table), f)
+
+            if self.cache_id is not None:
+                self._write_cache(cache_dir, cache_file, table)
+
         if table.sr_conflicts:
             warnings.warn(
-                "%d shift/reduce conflict%s" % (len(table.sr_conflicts), "s" if len(table.sr_conflicts) > 1 else ""),
+                "%d shift/reduce conflict%s" % (
+                    len(table.sr_conflicts),
+                    "s" if len(table.sr_conflicts) > 1 else ""
+                ),
                 ParserGeneratorWarning,
                 stacklevel=2,
             )
         if table.rr_conflicts:
             warnings.warn(
-                "%d reduce/reduce conflict%s" % (len(table.rr_conflicts), "s" if len(table.rr_conflicts) > 1 else ""),
+                "%d reduce/reduce conflict%s" % (
+                    len(table.rr_conflicts),
+                    "s" if len(table.rr_conflicts) > 1 else ""
+                ),
                 ParserGeneratorWarning,
                 stacklevel=2,
             )
         return LRParser(table, self.error_handler)
 
+    def _write_cache(self, cache_dir, cache_file, table):
+        if not os.path.exists(cache_dir):
+            try:
+                os.makedirs(cache_dir, mode=0o0700)
+            except OSError as e:
+                if e.errno == errno.EROFS:
+                    return
+                raise
+
+        with tempfile.NamedTemporaryFile(dir=cache_dir, delete=False, mode="w") as f:
+            json.dump(self.serialize_table(table), f)
+        os.rename(f.name, cache_file)
+
 
 def digraph(X, R, FP):
     N = dict.fromkeys(X, 0)
@@ -202,7 +266,8 @@ def traverse(x, N, stack, F, X, R, FP):
 
 
 class LRTable(object):
-    def __init__(self, grammar, lr_action, lr_goto, default_reductions, sr_conflicts, rr_conflicts):
+    def __init__(self, grammar, lr_action, lr_goto, default_reductions,
+                 sr_conflicts, rr_conflicts):
         self.grammar = grammar
         self.lr_action = lr_action
         self.lr_goto = lr_goto
@@ -282,7 +347,7 @@ class LRTable(object):
                                         chosenp, rejectp = oldp, pp
                                     rr_conflicts.append((st, repr(chosenp), repr(rejectp)))
                                 else:
-                                    raise LALRError("Unknown conflict in state %d" % st)
+                                    raise ParserGeneratorError("Unknown conflict in state %d" % st)
                             else:
                                 st_action[a] = -p.number
                                 st_actionp[a] = p
@@ -298,7 +363,7 @@ class LRTable(object):
                                 r = st_action[a]
                                 if r > 0:
                                     if r != j:
-                                        raise LALRError("Shift/shift conflict in state %d" % st)
+                                        raise ParserGeneratorError("Shift/shift conflict in state %d" % st)
                                 elif r < 0:
                                     rprec, rlevel = grammar.productions[st_actionp[a].number].prec
                                     sprec, slevel = grammar.precedence.get(a, ("right", 0))
@@ -312,7 +377,7 @@ class LRTable(object):
                                         if not slevel and not rlevel:
                                             sr_conflicts.append((st, repr(a), "reduce"))
                                 else:
-                                    raise LALRError("Unknown conflict in state %d" % st)
+                                    raise ParserGeneratorError("Unknown conflict in state %d" % st)
                             else:
                                 st_action[a] = j
                                 st_actionp[a] = p
@@ -442,15 +507,19 @@ class LRTable(object):
 
     @classmethod
     def compute_read_sets(cls, grammar, C, ntrans, nullable, add_count, cidhash, goto_cache):
-        FP = lambda x: cls.dr_relation(grammar, C, x, nullable, add_count, goto_cache)
-        R = lambda x: cls.reads_relation(C, x, nullable, add_count, cidhash, goto_cache)
-        return digraph(ntrans, R, FP)
+        return digraph(
+            ntrans,
+            R=lambda x: cls.reads_relation(C, x, nullable, add_count, cidhash, goto_cache),
+            FP=lambda x: cls.dr_relation(grammar, C, x, nullable, add_count, goto_cache)
+        )
 
     @classmethod
     def compute_follow_sets(cls, ntrans, readsets, includesets):
-        FP = lambda x: readsets[x]
-        R = lambda x: includesets.get(x, [])
-        return digraph(ntrans, R, FP)
+        return digraph(
+            ntrans,
+            R=lambda x: includesets.get(x, []),
+            FP=lambda x: readsets[x],
+        )
 
     @classmethod
     def dr_relation(cls, grammar, C, trans, nullable, add_count, goto_cache):
